"""This file contains keywords that occur in the OWL ontology config files."""

IDENTIFIER_KEY = "identifier"
RDF_FILE_KEY = "ontology_file"
NAMESPACES_KEY = "namespaces"
FILE_FORMAT_KEY = "format"
REQUIREMENTS_KEY = "requirements"

MANDATORY_KEYS = [IDENTIFIER_KEY, RDF_FILE_KEY, NAMESPACES_KEY]

ALL_KEYS = [
    IDENTIFIER_KEY,
    RDF_FILE_KEY,
    NAMESPACES_KEY,
    FILE_FORMAT_KEY,
    REQUIREMENTS_KEY,
]
